package port

import (
	"net/http"
	"strconv"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/port/internal/apimodel"
)

const apiTribeSnapshotSortMaxLength = 1

var apiTribeSnapshotSortAllowedValues = []domain.TribeSnapshotSort{
	domain.TribeSnapshotSortDateASC,
	domain.TribeSnapshotSortDateDESC,
}

//nolint:gocyclo
func (h *apiHTTPHandler) ListTribeTribeSnapshots(
	w http.ResponseWriter,
	r *http.Request,
	_ apimodel.VersionCodePathParam,
	serverKey apimodel.ServerKeyPathParam,
	tribeID apimodel.TribeIdPathParam,
	params apimodel.ListTribeTribeSnapshotsParams,
) {
	domainParams := domain.NewListTribeSnapshotsParams()

	if err := domainParams.SetSort([]domain.TribeSnapshotSort{domain.TribeSnapshotSortIDASC}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
		return
	}

	if err := domainParams.SetTribeIDs([]int{tribeID}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
		return
	}

	if params.Sort != nil {
		if err := domainParams.PrependSortString(
			*params.Sort,
			apiTribeSnapshotSortAllowedValues,
			apiTribeSnapshotSortMaxLength,
		); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
			return
		}
	} else {
		if err := domainParams.PrependSort([]domain.TribeSnapshotSort{domain.TribeSnapshotSortDateASC}); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
			return
		}
	}

	if err := domainParams.SetServerKeys([]string{serverKey}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
		return
	}

	if params.Limit != nil {
		if err := domainParams.SetLimit(*params.Limit); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
			return
		}
	}

	if params.Cursor != nil {
		if err := domainParams.SetEncodedCursor(*params.Cursor); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListTribeSnapshotsErrorPath).render(w, r, err)
			return
		}
	}

	res, err := h.tribeSnapshotSvc.ListWithRelations(r.Context(), domainParams)
	if err != nil {
		h.errorRenderer.render(w, r, err)
		return
	}

	renderJSON(w, r, http.StatusOK, apimodel.NewListTribeSnapshotsResponse(res))
}

func formatListTribeSnapshotsErrorPath(segments []domain.ErrorPathSegment) []string {
	if segments[0].Model != "ListTribeSnapshotsParams" {
		return nil
	}

	switch segments[0].Field {
	case "cursor":
		return []string{"$query", "cursor"}
	case "limit":
		return []string{"$query", "limit"}
	case "sort":
		path := []string{"$query", "sort"}
		if segments[0].Index >= 0 {
			path = append(path, strconv.Itoa(segments[0].Index))
		}
		return path
	default:
		return nil
	}
}
